#
# python_grabber
#
# Authors:
#  Andrea Schiavinato <andrea.schiavinato84@gmail.com>
#
# Copyright (C) 2019 Andrea Schiavinato
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from qgrabber.pygrabber.dshow_graph import *


class PyGrabber:
    def __init__(self, callback):
        self.graph = FilterGraph()
        self.callback = callback
        self.preview_graph_prepared = False
        self.recording_prepared = False

    def get_video_devices(self):
        return self.graph.get_input_devices()

    def get_audio_devices(self):
        return self.graph.get_audio_devices()

    def get_video_compressors(self):
        return self.graph.get_video_compressors()

    def get_audio_compressors(self):
        return self.graph.get_audio_compressors()

    def get_asf_profiles(self):
        return self.graph.get_asf_profiles()

    def set_device(self, input_device_index):
        self.graph.add_video_input_device(input_device_index)

    def start_preview(self, handle):
        if not self.preview_graph_prepared:
            if self.recording_prepared:
                self.graph.remove_all_filters_but_video_source()
                self.recording_prepared = False

            self.graph.add_sample_grabber(self.callback)
            self.graph.add_default_render()
            self.graph.prepare_preview_graph()
            self.graph.configure_render(handle)
            self.preview_graph_prepared = True
        self.graph.run()

    def start_recording(
        self,
        audio_device_index,
        video_compressor_index,
        audio_compressor_index,
        filename,
        handle,
    ):
        self.graph.stop()
        self.preview_graph_prepared = False
        self.graph.remove_all_filters_but_video_source()
        self.graph.add_default_render()
        if video_compressor_index is not None:
            self.graph.add_video_compressor(video_compressor_index)
        if audio_device_index is not None:
            self.graph.add_audio_input_device(audio_device_index)
            if audio_compressor_index is not None:
                self.graph.add_audio_compressor(audio_compressor_index)
        self.graph.add_file_writer_and_muxer(filename)
        self.graph.prepare_recording_graph()
        self.graph.configure_render(handle)
        self.recording_prepared = True
        self.graph.run()

    def stop(self):
        self.graph.stop()

    def update_window(self, width, height):
        self.graph.update_window(width, height)

    def set_device_properties(self):
        self.graph.get_input_device().set_properties()

    def display_format_dialog(self):
        self.graph.get_input_device().show_format_dialog()

    def grab_frame(self):
        self.graph.grab_frame()

    def get_status(self):
        graph_state = self.graph.get_state()
        device_name = self.graph.get_input_device().Name
        resolution = self.graph.get_input_device().get_current_format()
        if graph_state == StateGraph.Stopped:
            return "Stopped"
        elif graph_state == StateGraph.Running:
            return f"{'Recording' if self.graph.is_recording else 'Playing'} {device_name} [{resolution[0]}x{resolution[1]}]"
        elif graph_state == StateGraph.Paused:
            return f"Connected to {device_name} - paused"
        return None



#
# python_grabber
#
# Authors:
#  Andrea Schiavinato <andrea.schiavinato84@gmail.com>
#
# Copyright (C) 2019 Andrea Schiavinato
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

# see https://github.com/tpn/winsdk-10/blob/master/Include/10.0.16299.0/um/axextend.idl for interface spec.

from qgrabber.pygrabber.moniker import *
from qgrabber.pygrabber.win_common_types import *
from comtypes import *
from comtypes import client
from ctypes.wintypes import RECT, SIZE, ULONG, LPOLESTR, DWORD, LONG
from comtypes.automation import IDispatch
from ctypes import c_int, c_long, c_longlong

qedit = client.GetModule("qedit.dll")
quartz = client.GetModule("quartz.dll")


PIN_IN = 0
PIN_OUT = 1


class BITMAPINFOHEADER(Structure):
    _fields_ = [
        ("biSize", c_uint32),
        ("biWidth", c_int),
        ("biHeight", c_int),
        ("biPlanes", c_short),
        ("biBitCount", c_short),
        ("biCompression", c_uint32),
        ("biSizeImage", c_uint32),
        ("biXPelsPerMeter", c_long),
        ("biYPelsPerMeter", c_long),
        ("biClrUsed", c_uint32),
        ("biClrImportant", c_uint32),
    ]


class VIDEOINFOHEADER(Structure):
    _fields_ = (
        ("source", RECT),
        ("target", RECT),
        ("bit_rate", DWORD),
        ("bit_error_rate", DWORD),
        ("avg_time_per_frame", REFERENCE_TIME),
        ("bmi_header", BITMAPINFOHEADER),
    )


class VIDEO_STREAM_CONFIG_CAPS(Structure):
    _fields_ = (
        ("guid", GUID),
        ("VideoStandard", ULONG),
        ("InputSize", SIZE),
        ("MinCroppingSize", SIZE),
        ("MaxCroppingSize", SIZE),
        ("CropGranularityX", c_int),
        ("CropGranularityY", c_int),
        ("CropAlignX", c_int),
        ("CropAlignY", c_int),
        ("MinOutputSize", SIZE),
        ("MaxOutputSize", SIZE),
        ("OutputGranularityX", c_int),
        ("OutputGranularityX", c_int),
        ("StretchTapsX", c_int),
        ("StretchTapsY", c_int),
        ("ShrinkTapsX", c_int),
        ("ShrinkTapsY", c_int),
        ("MinFrameInterval", c_longlong),
        ("MaxFrameInterval", c_longlong),
        ("MinBitsPerSecond", LONG),
        ("MaxBitsPerSecond", LONG),
    )


class ISampleGrabber(IUnknown):
    _case_insensitive_ = True
    "ISampleGrabber Interface"
    _iid_ = GUID("{6B652FFF-11FE-4FCE-92AD-0266B5D7C78F}")
    _idlflags_ = []


ISampleGrabber._methods_ = [
    COMMETHOD([], HRESULT, "SetOneShot", ([], c_int, "OneShot")),
    COMMETHOD([], HRESULT, "SetMediaType", ([], POINTER(qedit._AMMediaType), "pType")),
    COMMETHOD(
        [],
        HRESULT,
        "GetConnectedMediaType",
        (["out"], POINTER(qedit._AMMediaType), "pType"),
    ),
    COMMETHOD([], HRESULT, "SetBufferSamples", ([], c_int, "BufferThem")),
    COMMETHOD(
        [],
        HRESULT,
        "GetCurrentBuffer",
        (["in", "out"], POINTER(c_int), "pBufferSize"),
        (["out"], POINTER(c_int), "pBuffer"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetCurrentSample",
        (["out", "retval"], POINTER(POINTER(qedit.IMediaSample)), "ppSample"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "SetCallback",
        ([], POINTER(qedit.ISampleGrabberCB), "pCallback"),
        ([], c_int, "WhichMethodToCallback"),
    ),
]


class ICreateDevEnum(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{29840822-5B84-11D0-BD3B-00A0C911CE86}")
    _idlflags_ = []


ICreateDevEnum._methods_ = [
    COMMETHOD(
        [],
        HRESULT,
        "CreateClassEnumerator",
        (["in"], POINTER(GUID), "clsidDeviceClass"),
        (["out"], POINTER(POINTER(IEnumMoniker)), "ppEnumMoniker"),
        (["in"], c_int, "dwFlags"),
    )
]


class IAMStreamConfig(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{c6e13340-30ac-11d0-a18c-00a0c9118956}")
    _idlflags_ = []


IAMStreamConfig._methods_ = [
    COMMETHOD([], HRESULT, "SetFormat", (["in"], POINTER(qedit._AMMediaType), "pmt")),
    COMMETHOD(
        [], HRESULT, "GetFormat", (["out"], POINTER(POINTER(qedit._AMMediaType)), "pmt")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetNumberOfCapabilities",
        (["out"], POINTER(c_int), "piCount"),
        (["out"], POINTER(c_int), "piSize"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetStreamCaps",  # https://docs.microsoft.com/en-us/previous-versions/ms784114(v%3Dvs.85)
        (["in"], c_int, "iIndex"),
        (["out"], POINTER(POINTER(qedit._AMMediaType)), "pmt"),
        (["out"], POINTER(VIDEO_STREAM_CONFIG_CAPS), "pSCC"),
    ),
]


class ISpecifyPropertyPages(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{B196B28B-BAB4-101A-B69C-00AA00341D07}")
    _idlflags_ = []


class CAUUID(Structure):
    _fields_ = (
        ("element_count", ULONG),
        ("elements", POINTER(GUID)),
    )


ISpecifyPropertyPages._methods_ = [
    COMMETHOD([], HRESULT, "GetPages", (["out"], POINTER(CAUUID), "pPages"),)
]


class IVideoWindow(IDispatch):
    _case_insensitive_ = True
    _iid_ = GUID("{56a868b4-0ad4-11ce-b03a-0020af0ba770}")
    _idlflags_ = []


IVideoWindow._methods_ = [
    COMMETHOD([], HRESULT, "put_Caption", (["in"], BSTR, "strCaption")),
    COMMETHOD(
        [], HRESULT, "get_Caption", (["retval", "out"], POINTER(BSTR), "strCaption")
    ),
    COMMETHOD([], HRESULT, "put_WindowStyle", (["in"], c_long, "WindowStyle")),
    COMMETHOD(
        [],
        HRESULT,
        "get_WindowStyle",
        (["retval", "out"], POINTER(c_long), "WindowStyle"),
    ),
    COMMETHOD([], HRESULT, "put_WindowStyleEx", (["in"], c_long, "WindowStyleEx")),
    COMMETHOD(
        [],
        HRESULT,
        "get_WindowStyleEx",
        (["retval", "out"], POINTER(c_long), "WindowStyleEx"),
    ),
    COMMETHOD([], HRESULT, "put_AutoShow", (["in"], c_long, "AutoShow")),
    COMMETHOD(
        [], HRESULT, "get_AutoShow", (["retval", "out"], POINTER(c_long), "AutoShow")
    ),
    COMMETHOD([], HRESULT, "put_WindowState", (["in"], c_long, "WindowState")),
    COMMETHOD(
        [],
        HRESULT,
        "get_WindowState",
        (["retval", "out"], POINTER(c_long), "WindowState"),
    ),
    COMMETHOD(
        [], HRESULT, "put_BackgroundPalette", (["in"], c_long, "BackgroundPalette")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "get_BackgroundPalette",
        (["retval", "out"], POINTER(c_long), "pBackgroundPalette"),
    ),
    COMMETHOD([], HRESULT, "put_Visible", (["in"], c_long, "Visible")),
    COMMETHOD(
        [], HRESULT, "get_Visible", (["retval", "out"], POINTER(c_long), "pVisible")
    ),
    COMMETHOD([], HRESULT, "put_Left", (["in"], c_long, "Left")),
    COMMETHOD([], HRESULT, "get_Left", (["retval", "out"], POINTER(c_long), "pLeft")),
    COMMETHOD([], HRESULT, "put_Width", (["in"], c_long, "Width")),
    COMMETHOD([], HRESULT, "get_Width", (["retval", "out"], POINTER(c_long), "pWidth")),
    COMMETHOD([], HRESULT, "put_Top", (["in"], c_long, "Top")),
    COMMETHOD([], HRESULT, "get_Top", (["retval", "out"], POINTER(c_long), "pTop")),
    COMMETHOD([], HRESULT, "put_Height", (["in"], c_long, "Height")),
    COMMETHOD(
        [], HRESULT, "get_Height", (["retval", "out"], POINTER(c_long), "pHeight")
    ),
    COMMETHOD([], HRESULT, "put_Owner", (["in"], OLE_HANDLE, "Owner")),
    COMMETHOD(
        [], HRESULT, "get_Owner", (["retval", "out"], POINTER(OLE_HANDLE), "Owner")
    ),
    COMMETHOD([], HRESULT, "put_MessageDrain", (["in"], OLE_HANDLE, "Drain")),
    COMMETHOD(
        [],
        HRESULT,
        "get_MessageDrain",
        (["retval", "out"], POINTER(OLE_HANDLE), "Drain"),
    ),
    COMMETHOD(
        [], HRESULT, "get_BorderColor", (["retval", "out"], POINTER(c_long), "Color")
    ),
    COMMETHOD([], HRESULT, "put_BorderColor", (["in"], c_long, "Color")),
    COMMETHOD(
        [],
        HRESULT,
        "get_FullScreenMode",
        (["retval", "out"], POINTER(c_long), "FullScreenMode"),
    ),
    COMMETHOD([], HRESULT, "put_FullScreenMode", (["in"], c_long, "FullScreenMode")),
    COMMETHOD([], HRESULT, "SetWindowForeground", (["in"], c_long, "Focus")),
    COMMETHOD(
        [],
        HRESULT,
        "NotifyOwnerMessage",
        (["in"], OLE_HANDLE, "hwnd"),
        (["in"], c_long, "uMsg"),
        (["in"], LONG_PTR, "wParam"),
        (["in"], LONG_PTR, "lParam"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "SetWindowPosition",
        (["in"], c_long, "Left"),
        (["in"], c_long, "Top"),
        (["in"], c_long, "Width"),
        (["in"], c_long, "Height"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetWindowPosition",
        (["out"], POINTER(c_long), "pLeft"),
        (["out"], POINTER(c_long), "pTop"),
        (["out"], POINTER(c_long), "pWidth"),
        (["out"], POINTER(c_long), "pHeight"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetMinIdealImageSize",
        (["out"], POINTER(c_long), "pWidth"),
        (["out"], POINTER(c_long), "pHeight"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetMaxIdealImageSize",
        (["out"], POINTER(c_long), "pWidth"),
        (["out"], POINTER(c_long), "pHeight"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetRestorePosition",
        (["out"], POINTER(c_long), "pLeft"),
        (["out"], POINTER(c_long), "pTop"),
        (["out"], POINTER(c_long), "pWidth"),
        (["out"], POINTER(c_long), "pHeight"),
    ),
    COMMETHOD([], HRESULT, "HideCursor", (["in"], c_long, "HideCursor")),
    COMMETHOD(
        [], HRESULT, "IsCursorHidden", (["out"], POINTER(c_long), "CursorHidden")
    ),
]


class ICaptureGraphBuilder2(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{93E5A4E0-2D50-11d2-ABFA-00A0C9C6E38D}")
    _idlflags_ = []


ICaptureGraphBuilder2._methods_ = [
    COMMETHOD(
        [], HRESULT, "SetFiltergraph", (["in"], POINTER(qedit.IFilterGraph), "pfg")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetFiltergraph",
        (["out"], POINTER(POINTER(qedit.IFilterGraph)), "pfg"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "SetOutputFileName",
        (["in"], POINTER(GUID), "pType"),
        (["in"], LPCOLESTR, "lpstrFile"),
        (["out"], POINTER(POINTER(qedit.IBaseFilter)), "ppf"),
        (["out"], POINTER(POINTER(qedit.IBaseFilter)), "pSink"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "FindInterface",
        (["in"], POINTER(GUID), "pCategory"),
        (["in"], POINTER(GUID), "pType"),
        (["in"], POINTER(qedit.IBaseFilter), "pf"),
        (["in"], REFIID, "REFIID"),
        (["out"], POINTER(POINTER(None)), "ppINT"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "RenderStream",
        (["in"], POINTER(GUID), "pCategory"),
        (["in"], POINTER(GUID), "pType"),
        (["in"], POINTER(qedit.IBaseFilter), "pSource"),
        (["in"], POINTER(qedit.IBaseFilter), "pIntermediate"),
        (["in"], POINTER(qedit.IBaseFilter), "pSink"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "ControlStream",
        (["in"], POINTER(GUID), "pCategory"),
        (["in"], POINTER(GUID), "pType"),
        (["in"], POINTER(qedit.IBaseFilter), "pFilter"),
        (["in"], POINTER(REFERENCE_TIME), "pstart"),
        (["in"], POINTER(REFERENCE_TIME), "pstop"),
        (["in"], WORD, "wStartCookie"),
        (["in"], WORD, "wStopCookie"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "AllocCapFile",
        (["in"], LPCOLESTR, "lpstr"),
        (["in"], DWORDLONG, "dwlSize"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "CopyCaptureFile",
        (["in"], LPOLESTR, "lpwstrOld"),
        (["in"], LPOLESTR, "lpwstrNew"),
        (["in"], c_int, "fAllowEscAbort"),
        (["in"], POINTER(IUnknown), "pCallback"),
    ),
    # (['in'], POINTER(IAMCopyCaptureFileProgress), 'pCallback'),
    COMMETHOD(
        [],
        HRESULT,
        "FindPin",
        (["in"], POINTER(qedit.IBaseFilter), "pSource"),
        (["in"], c_int, "pindir"),
        (["in"], POINTER(GUID), "pCategory"),
        (["in"], POINTER(GUID), "pType"),
        (["in"], c_int, "fUnconnected"),
        (["in"], c_int, "num"),
        (["out"], POINTER(POINTER(qedit.IPin)), "ppPin"),
    ),
]



import asyncio
import logging
from typing import List
import wx
from aio_wx_widgets.frame import DefaultFrame
from aio_wx_widgets.widgets import button
from aio_wx_widgets.widgets import text
from pyzbar.pyzbar import Decoded
from wx.core import wxdate2pydate
from wxasync import WxAsyncApp

from qgrabber.controller import Controller
from qgrabber.model import ScannerModel
from qgrabber.view import ScannerView

_LOGGER = logging.getLogger(__name__)


class MainWindow(DefaultFrame):
    def __init__(self):
        super().__init__("Main window")
        model = ScannerModel(stop_on_scan=True)
        # model.on_code_scanned.append(self._on_scan)

        view_width = 640
        view_height = 480

        self.controller = Controller(
            model, view_width=view_width, view_height=view_height
        )
        view = ScannerView(self, self.controller, width=view_width, height=view_height)

        self.add(view, layout=wx.ALIGN_CENTER, create=False)

        self.add(button.async_button("start", self._on_start))
        self.add(button.async_button("stop", self._on_stop))
        self.scan_results: text.Text = self.add(text.Text("unknown"))

    def _on_scan(self, result: List[Decoded]):
        first_result = result[0]
        self.scan_results.set_text(str(first_result.data))

    async def _on_start(self, event):
        result = await self.controller.scan_code()
        # self.controller.start_scan()
        self.scan_results.set_text(str(result))

    async def _on_stop(self, event):
        print("stopping")


async def run_app():
    app = WxAsyncApp()
    main_window = MainWindow()
    main_window.Show()
    app.SetTopWindow(main_window)
    await app.MainLoop()


if __name__ == "__main__":
    # os.environ["DEBUGGING"] = "1"
    logging.basicConfig(level=logging.DEBUG)
    loop = asyncio.get_event_loop()
    # app = WxAsyncApp()
    # main_window = MainWindow()
    # main_window.Show()
    # app.SetTopWindow(main_window)
    # loop.run_until_complete(app.MainLoop())
    loop.run_until_complete(run_app())



#
# python_grabber
#
# Authors:
#  Andrea Schiavinato <andrea.schiavinato84@gmail.com>
#
# Copyright (C) 2019 Andrea Schiavinato
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import numpy as np
import os.path
from enum import Enum
from comtypes.persist import IPropertyBag

from qgrabber.pygrabber.dshow_core import *
from qgrabber.pygrabber.windows_media import *
from qgrabber.pygrabber.dshow_ids import *
from qgrabber.pygrabber.win_api_extra import *


class StateGraph(Enum):
    Stopped = 0
    Paused = 1
    Running = 2


class RecordingFormat(Enum):
    AVI = 0
    ASF = 1


class FilterType(Enum):
    video_input = 0
    audio_input = 1
    video_compressor = 2
    audio_compressor = 3
    sample_grabber = 4
    render = 5
    file_sink = 6
    muxer = 7
    smart_tee = 8


class Filter:
    # Wrapper around a Direct Show filter
    def __init__(self, instance, name, capture_builder):
        self.instance = instance
        self.capture_builder = capture_builder
        self.Name = name
        self.out_pins = []
        self.in_pins = []
        self.reload_pins()

    def get_out(self):
        return self.out_pins[0]

    def get_in(self, index=0):
        return self.in_pins[index]

    def find_pin(self, direction, category=None, type=None, unconnected=True):
        try:
            return self.capture_builder.FindPin(
                self.instance, direction, category, type, unconnected, 0
            )
        except COMError:
            return None  # assuming preview pin not found

    def reload_pins(self):
        # 0 = in, 1 = out
        self.out_pins = []
        self.in_pins = []
        enum = self.instance.EnumPins()
        pin, count = enum.Next(1)
        while count > 0:
            if pin.QueryDirection() == 0:
                self.in_pins.append(pin)
            else:
                self.out_pins.append(pin)
            pin, count = enum.Next(1)

    def set_properties(self):
        show_properties(self.instance)

    def get_name(self):
        filter_info = self.instance.QueryFilterInfo()
        return wstring_at(filter_info.achName)

    def print_info(self):
        print(f"Pins of: {self.get_name()}")
        enum = self.instance.EnumPins()
        pin, count = enum.Next(1)
        while count > 0:
            info = pin.QueryPinInfo()
            direction, name = (info.dir, wstring_at(info.achName))
            print(f"PIN {'in' if direction == 0 else 'out'} - {name}")
            pin, count = enum.Next(1)


class VideoInput(Filter):
    def __init__(self, args, capture_builder):
        Filter.__init__(self, args[0], args[1], capture_builder)

    def get_current_format(self):
        stream_config = self.get_out().QueryInterface(IAMStreamConfig)
        media_type = stream_config.GetFormat()
        p_video_info_header = cast(
            media_type.contents.pbFormat, POINTER(VIDEOINFOHEADER)
        )
        bmp_header = p_video_info_header.contents.bmi_header
        return bmp_header.biWidth, bmp_header.biHeight

    def get_formats(self):
        # https://docs.microsoft.com/en-us/windows/win32/directshow/configure-the-video-output-format
        stream_config = self.get_out().QueryInterface(IAMStreamConfig)
        media_types_count, _ = stream_config.GetNumberOfCapabilities()
        result = []
        for i in range(0, media_types_count):
            media_type, capability = stream_config.GetStreamCaps(i)
            p_video_info_header = cast(
                media_type.contents.pbFormat, POINTER(VIDEOINFOHEADER)
            )
            bmp_header = p_video_info_header.contents.bmi_header
            result.append(
                {
                    "index": i,
                    "media_type_str": subtypes[str(media_type.contents.subtype)],
                    "width": bmp_header.biWidth,
                    "height": bmp_header.biHeight,
                    "min_framerate": 10000000 / capability.MinFrameInterval,
                    "max_framerate": 10000000 / capability.MaxFrameInterval,
                }
            )
            # print(f"{capability.MinOutputSize.cx}x{capability.MinOutputSize.cx} - {capability.MaxOutputSize.cx}x{capability.MaxOutputSize.cx}")
        return result

    def set_format(self, format_index):
        stream_config = self.get_out().QueryInterface(IAMStreamConfig)
        media_type, _ = stream_config.GetStreamCaps(format_index)
        stream_config.SetFormat(media_type)

    def show_format_dialog(self):
        show_properties(self.get_out())


class AudioInput(Filter):
    def __init__(self, args, capture_builder):
        Filter.__init__(self, args[0], args[1], capture_builder)


class VideoCompressor(Filter):
    def __init__(self, args, capture_builder):
        Filter.__init__(self, args[0], args[1], capture_builder)


class AudioCompressor(Filter):
    def __init__(self, args, capture_builder):
        Filter.__init__(self, args[0], args[1], capture_builder)


class Render(Filter):
    def __init__(self, instance, capture_builder):
        Filter.__init__(self, instance, "Render", capture_builder)
        try:
            self.video_window = self.instance.QueryInterface(IVideoWindow)
        except COMError:
            self.video_window = None  # probably interface IVideoWindow not supported because using NullRender

    def configure_video_window(self, handle):
        # must be called after the graph is connected
        self.video_window.put_Owner(handle)
        self.video_window.put_WindowStyle(WS_CHILD | WS_CLIPSIBLINGS)

    def set_window_position(self, x, y, width, height):
        self.video_window.SetWindowPosition(x, y, width, height)


class SampleGrabber(Filter):
    def __init__(self, capture_builder):
        Filter.__init__(
            self,
            client.CreateObject(
                GUID(clsids.CLSID_SampleGrabber), interface=qedit.IBaseFilter
            ),
            "Sample Grabber",
            capture_builder,
        )
        self.sample_grabber = self.instance.QueryInterface(ISampleGrabber)
        self.callback = None

    def set_callback(self, callback, which_method_to_callback):
        self.callback = callback
        self.sample_grabber.SetCallback(callback, which_method_to_callback)

    def set_media_type(self, media_type, media_subtype):
        sg_type = qedit._AMMediaType()
        sg_type.majortype = GUID(media_type)
        sg_type.subtype = GUID(media_subtype)
        self.sample_grabber.SetMediaType(sg_type)

    def get_resolution(self):
        media_type = self.sample_grabber.GetConnectedMediaType()
        p_video_info_header = cast(media_type.pbFormat, POINTER(VIDEOINFOHEADER))
        bmp_header = p_video_info_header.contents.bmi_header
        return bmp_header.biWidth, bmp_header.biHeight

    def initialize_after_connection(self):
        self.callback.image_resolution = self.get_resolution()


class SmartTee(Filter):
    def __init__(self, capture_builder):
        Filter.__init__(
            self,
            client.CreateObject(
                GUID(clsids.CLSID_SmartTee), interface=qedit.IBaseFilter
            ),
            "Smart Tee",
            capture_builder,
        )


class Muxer(Filter):
    def __init__(self, args, capture_builder):
        Filter.__init__(self, args, "Muxer", capture_builder)


class SystemDeviceEnum:
    def __init__(self):
        self.system_device_enum = client.CreateObject(
            clsids.CLSID_SystemDeviceEnum, interface=ICreateDevEnum
        )

    def get_available_filters(self, category_clsid):
        filter_enumerator = self.system_device_enum.CreateClassEnumerator(
            GUID(category_clsid), dwFlags=0
        )
        moniker, count = filter_enumerator.Next(1)
        result = []
        while count > 0:
            result.append(get_moniker_name(moniker))
            moniker, count = filter_enumerator.Next(1)
        return result

    def get_filter_by_index(self, category_clsid, index):
        filter_enumerator = self.system_device_enum.CreateClassEnumerator(
            GUID(category_clsid), dwFlags=0
        )
        moniker, count = filter_enumerator.Next(1)
        i = 0
        while i != index and count > 0:
            moniker, count = filter_enumerator.Next(1)
            i = i + 1

        return (
            moniker.BindToObject(0, 0, qedit.IBaseFilter._iid_).QueryInterface(
                qedit.IBaseFilter
            ),
            get_moniker_name(moniker),
        )


class FilterFactory:
    def __init__(self, system_device_enum, capture_builder):
        self.system_device_enum = system_device_enum
        self.capture_builder = capture_builder

    def build_filter(self, filter_type, id):
        if filter_type == FilterType.video_input:
            return VideoInput(
                self.system_device_enum.get_filter_by_index(
                    DeviceCategories.VideoInputDevice, id
                ),
                self.capture_builder,
            )
        elif filter_type == FilterType.audio_input:
            return AudioInput(
                self.system_device_enum.get_filter_by_index(
                    DeviceCategories.AudioInputDevice, id
                ),
                self.capture_builder,
            )
        elif filter_type == FilterType.video_compressor:
            return VideoCompressor(
                self.system_device_enum.get_filter_by_index(
                    DeviceCategories.VideoCompressor, id
                ),
                self.capture_builder,
            )
        elif filter_type == FilterType.audio_compressor:
            return AudioCompressor(
                self.system_device_enum.get_filter_by_index(
                    DeviceCategories.AudioCompressor, id
                ),
                self.capture_builder,
            )
        elif filter_type == FilterType.render:
            return Render(
                client.CreateObject(GUID(id), interface=qedit.IBaseFilter),
                self.capture_builder,
            )
        elif filter_type == FilterType.sample_grabber:
            return SampleGrabber(self.capture_builder)
        elif filter_type == FilterType.muxer:
            return Muxer(id, self.capture_builder)
        elif filter_type == FilterType.smart_tee:
            return SmartTee(self.capture_builder)
        else:
            raise ValueError("Cannot create filter", filter_type, id)


class MediaType:
    def __init__(self, majortype_guid, subtype_guid):
        self.instance = qedit._AMMediaType()
        self.instance.majortype = GUID(majortype_guid)
        self.instance.subtype = GUID(subtype_guid)


class WmProfileManager:
    def __init__(self):
        self.profile_manager = POINTER(IWMProfileManager2)()
        WMCreateProfileManager(byref(self.profile_manager))
        self.profile_manager.SetSystemProfileVersion(0x00080000)
        self.profiles, self.profiles_names = self.__load_profiles()

    def __load_profiles(self):
        nr_profiles = self.profile_manager.GetSystemProfileCount()
        profiles = [
            self.profile_manager.LoadSystemProfile(i) for i in range(0, nr_profiles)
        ]
        profiles_names = []
        buf = create_unicode_buffer(200)
        for profile in profiles:
            i = DWORD(200)
            profile.GetName(buf, pointer(i))
            profiles_names.append(buf.value)
        return profiles, profiles_names


class FilterGraph:
    def __init__(self):
        self.filter_graph = client.CreateObject(
            clsids.CLSID_FilterGraph, interface=qedit.IFilterGraph
        )
        self.graph_builder = self.filter_graph.QueryInterface(qedit.IGraphBuilder)
        self.media_control = self.filter_graph.QueryInterface(quartz.IMediaControl)
        self.media_event = self.filter_graph.QueryInterface(quartz.IMediaEvent)
        self.capture_builder = client.CreateObject(
            clsids.CLSID_CaptureGraphBuilder2, interface=ICaptureGraphBuilder2
        )
        self.capture_builder.SetFiltergraph(self.filter_graph)

        self.system_device_enum = SystemDeviceEnum()
        self.filter_factory = FilterFactory(
            self.system_device_enum, self.capture_builder
        )
        self.wm_profile_manager = WmProfileManager()

        self.filters = {}
        self.recording_format = None
        self.is_recording = False

    def __add_filter(self, filter_type, filter_id):
        assert not (filter_type in self.filters)
        filter = self.filter_factory.build_filter(filter_type, filter_id)
        self.filters[filter_type] = filter
        self.filter_graph.AddFilter(filter.instance, filter.Name)

    def add_video_input_device(self, index):
        self.__add_filter(FilterType.video_input, index)

    def add_audio_input_device(self, index):
        self.__add_filter(FilterType.audio_input, index)

    def add_video_compressor(self, index):
        self.__add_filter(FilterType.video_compressor, index)

    def add_audio_compressor(self, index):
        self.__add_filter(FilterType.audio_compressor, index)

    def add_sample_grabber(self, callback):
        self.__add_filter(FilterType.sample_grabber, None)
        sample_grabber = self.filters[FilterType.sample_grabber]
        sample_grabber_cb = SampleGrabberCallback(callback)
        sample_grabber.set_callback(sample_grabber_cb, 1)
        sample_grabber.set_media_type(MediaTypes.Video, MediaSubtypes.RGB24)

    def add_null_render(self):
        self.__add_filter(FilterType.render, clsids.CLSID_NullRender)

    def add_default_render(self):
        self.__add_filter(FilterType.render, clsids.CLSID_VideoRendererDefault)

    def add_video_mixing_render(self):
        self.__add_filter(FilterType.render, clsids.CLSID_VideoMixingRenderer)

    def add_file_writer_and_muxer(self, filename):
        extension = os.path.splitext(filename)[1].upper()
        mediasubtype = MediaSubtypes.ASF if extension == ".WMV" else MediaSubtypes.AVI
        self.recording_format = (
            RecordingFormat.ASF if extension == ".WMV" else RecordingFormat.AVI
        )
        mux, filesink = self.capture_builder.SetOutputFileName(
            GUID(mediasubtype), filename
        )
        self.filters[FilterType.muxer] = self.filter_factory.build_filter(
            FilterType.muxer, mux
        )

    def configure_asf_compressor(self):
        pass
        # asf_config = self.mux.QueryInterface(IConfigAsfWriter)
        # print(asf_config.GetCurrentProfileGuid())
        # profile = asf_config.GetCurrentProfile()

    def prepare_preview_graph(self):
        assert FilterType.video_input in self.filters
        assert FilterType.render in self.filters
        if FilterType.sample_grabber not in self.filters:
            self.graph_builder.Connect(
                self.filters[FilterType.video_input].get_out(),
                self.filters[FilterType.render].get_in(),
            )
        else:
            self.graph_builder.Connect(
                self.filters[FilterType.video_input].get_out(),
                self.filters[FilterType.sample_grabber].get_in(),
            )
            self.graph_builder.Connect(
                self.filters[FilterType.sample_grabber].get_out(),
                self.filters[FilterType.render].get_in(),
            )
            self.filters[FilterType.sample_grabber].initialize_after_connection()
        self.is_recording = False

    def __get_capture_and_preview_pins(self):
        preview_pin = self.filters[FilterType.video_input].find_pin(
            PIN_OUT, category=GUID(PinCategory.Preview)
        )
        capture_pin = self.filters[FilterType.video_input].find_pin(
            PIN_OUT, category=GUID(PinCategory.Capture)
        )

        if (preview_pin is None) or (capture_pin is None):
            self.__add_filter(FilterType.smart_tee, None)
            smart_tee = self.filters[FilterType.smart_tee]
            self.graph_builder.Connect(
                capture_pin if capture_pin is not None else preview_pin,
                smart_tee.get_in(),
            )
            # assuming the 1st output pin of the smart tee filter is always the capture one
            capture_pin, preview_pin = smart_tee.out_pins

        return preview_pin, capture_pin

    def prepare_recording_graph(self):
        #  in theory we could use self.capture_builder.RenderStream,
        #  but it is not working when including the video compressor :-(
        assert FilterType.video_input in self.filters
        assert FilterType.render in self.filters
        assert FilterType.muxer in self.filters

        preview_pin, capture_pin = self.__get_capture_and_preview_pins()

        if self.recording_format == RecordingFormat.ASF:
            self.graph_builder.Connect(
                capture_pin, self.filters[FilterType.muxer].get_in(1)
            )
            self.graph_builder.Connect(
                self.filters[FilterType.audio_input].get_out(),
                self.filters[FilterType.muxer].get_in(0),
            )
            self.graph_builder.Connect(
                preview_pin, self.filters[FilterType.render].get_in()
            )

        else:
            self.graph_builder.Connect(
                capture_pin, self.filters[FilterType.video_compressor].get_in()
            )
            self.graph_builder.Connect(
                self.filters[FilterType.video_compressor].get_out(),
                self.filters[FilterType.muxer].get_in(),
            )
            self.graph_builder.Connect(
                preview_pin, self.filters[FilterType.render].get_in()
            )

            if FilterType.audio_input in self.filters:
                self.graph_builder.Connect(
                    self.filters[FilterType.audio_input].get_out(),
                    self.filters[FilterType.audio_compressor].get_in(),
                )
                self.filters[FilterType.muxer].reload_pins()
                # when you connect an input pin of the muxer, an additional input pin is added
                self.graph_builder.Connect(
                    self.filters[FilterType.audio_compressor].get_out(),
                    self.filters[FilterType.muxer].get_in(1),
                )

        self.is_recording = True

    def configure_render(self, handle):
        self.filters[FilterType.render].configure_video_window(handle)

    def update_window(self, width, height):
        if FilterType.render in self.filters:
            img_w, img_h = self.filters[FilterType.video_input].get_current_format()
            scale_w = width / img_w
            scale_h = height / img_h
            scale = min(scale_w, scale_h, 1)
            self.filters[FilterType.render].set_window_position(
                0, 0, int(img_w * scale), int(img_h * scale)
            )

    def run(self):
        self.media_control.Run()

    def stop(self):
        if self.media_control is not None:
            # calling stop without calling prepare
            self.media_control.Stop()
        # if self.video_window is not None:
        # self.video_window.put_Visible(False)
        # self.video_window.put_Owner(0)

    def pause(self):
        self.media_control.Pause()

    def get_state(self):
        return StateGraph(
            self.media_control.GetState(0xFFFFFFFF)
        )  # 0xFFFFFFFF = infinite timeout

    def get_input_devices(self):
        return self.system_device_enum.get_available_filters(
            DeviceCategories.VideoInputDevice
        )

    def get_audio_devices(self):
        return self.system_device_enum.get_available_filters(
            DeviceCategories.AudioInputDevice
        )

    def get_video_compressors(self):
        return self.system_device_enum.get_available_filters(
            DeviceCategories.VideoCompressor
        )

    def get_audio_compressors(self):
        return self.system_device_enum.get_available_filters(
            DeviceCategories.AudioCompressor
        )

    def get_asf_profiles(self):
        return self.wm_profile_manager.profiles_names

    def grab_frame(self):
        if FilterType.sample_grabber in self.filters:
            self.filters[FilterType.sample_grabber].callback.grab_frame()
            return True
        else:
            return False

    def get_input_device(self):
        return self.filters[FilterType.video_input]

    def remove_filters(self):
        enum_filters = self.filter_graph.EnumFilters()
        filt, count = enum_filters.Next(1)
        while count > 0:
            self.filter_graph.RemoveFilter(filt)
            enum_filters.Reset()
            filt, count = enum_filters.Next(1)
        self.filters = {}

    def remove_all_filters_but_video_source(self):
        video_input = self.filters[FilterType.video_input]
        enum_filters = self.filter_graph.EnumFilters()
        filters_to_delete = []
        filt, count = enum_filters.Next(1)
        while count > 0:
            if filt != video_input.instance:
                filters_to_delete.append(filt)
            filt, count = enum_filters.Next(1)
        for filt in filters_to_delete:
            self.filter_graph.RemoveFilter(filt)
        self.filters = {FilterType.video_input: video_input}

    def print_debug_info(self):
        helper = FilterGraphDebugHelper(self.filter_graph)
        helper.print_graph_info()


class FilterGraphDebugHelper:
    def __init__(self, filter_graph):
        self.filter_graph = filter_graph

    def print_graph_info(self):
        enum_filters = self.filter_graph.EnumFilters()
        filt, count = enum_filters.Next(1)
        while count > 0:
            filterName = self.get_filter_name(filt)
            print(f"FILTER {filterName} [{filt}]")

            enum_pins = filt.EnumPins()
            pin, count = enum_pins.Next(1)
            while count > 0:
                pin_name, direction, connected_pin, owner = self.get_pin_info(pin)
                if connected_pin is not None:
                    connected_pin_name, _, _, connected_filter = self.get_pin_info(
                        connected_pin
                    )
                    connected_filter_name = self.get_filter_name(connected_filter)

                print(
                    f" - PIN {pin_name} {'in' if direction == 0 else 'out'} - Connected to: {connected_filter_name} [{pin}]"
                )

                pin, count = enum_pins.Next(1)
            filt, count = enum_filters.Next(1)

    def get_filter_name(self, filter):
        filter_info = filter.QueryFilterInfo()
        return wstring_at(filter_info.achName)

    def get_pin_info(self, pin):
        info = pin.QueryPinInfo()
        name = wstring_at(info.achName)
        owner_filter = info.pFilter
        try:
            connected_pin = pin.ConnectedTo()
        except:
            connected_pin = None
        return name, info.dir, connected_pin, owner_filter


class SampleGrabberCallback(COMObject):
    _com_interfaces_ = [qedit.ISampleGrabberCB]

    def __init__(self, callback):
        self.callback = callback
        self.cnt = 0
        self.keep_photo = False
        self.image_resolution = None
        super(SampleGrabberCallback, self).__init__()

    def grab_frame(self):
        self.keep_photo = True

    def SampleCB(self, this, SampleTime, pSample):
        return 0

    def BufferCB(self, this, SampleTime, pBuffer, BufferLen):
        if self.keep_photo:
            self.keep_photo = False
            img = np.ctypeslib.as_array(
                pBuffer, shape=(self.image_resolution[1], self.image_resolution[0], 3)
            )
            img = np.flip(np.copy(img), axis=0)
            self.callback(img)
        return 0

    # ALTERNATIVE
    # def BufferCB(self, this, SampleTime, pBuffer, BufferLen):
    #     if self.keep_photo:
    #         self.keep_photo = False
    #         bsize = self.image_resolution[1] *self.image_resolution[0] * 3
    #         img = pBuffer[:bsize]
    #         img = np.reshape(img, (self.image_resolution[1], self.image_resolution[0], 3))
    #         img = np.flip(img, axis=0)
    #         self.callback(img)
    #     return 0


def get_moniker_name(moniker):
    property_bag = moniker.BindToStorage(0, 0, IPropertyBag._iid_).QueryInterface(
        IPropertyBag
    )
    return property_bag.Read("FriendlyName", pErrorLog=None)


def show_properties(object):
    try:
        spec_pages = object.QueryInterface(ISpecifyPropertyPages)
        cauuid = spec_pages.GetPages()
        if cauuid.element_count > 0:
            whandle = windll.user32.GetTopWindow(None)
            OleCreatePropertyFrame(
                whandle,
                0,
                0,
                None,
                1,
                byref(cast(object, LPUNKNOWN)),
                cauuid.element_count,
                cauuid.elements,
                0,
                0,
                None,
            )
            windll.ole32.CoTaskMemFree(cauuid.elements)
    except COMError:
        pass



#
# python_grabber
#
# Authors:
#  Andrea Schiavinato <andrea.schiavinato84@gmail.com>
#
# Copyright (C) 2019 Andrea Schiavinato
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from ctypes import POINTER, HRESULT
from ctypes import windll
from ctypes.wintypes import DWORD, ULONG, HWND, UINT, LPCOLESTR, LCID, LPVOID

from comtypes import IUnknown, GUID

LPUNKNOWN = POINTER(IUnknown)
CLSID = GUID
LPCLSID = POINTER(CLSID)

WS_CHILD = 0x40000000
WS_CLIPSIBLINGS = 0x04000000

OleCreatePropertyFrame = windll.oleaut32.OleCreatePropertyFrame
OleCreatePropertyFrame.restype = HRESULT
OleCreatePropertyFrame.argtypes = (
    HWND,  # [in] hwndOwner
    UINT,  # [in] x
    UINT,  # [in] y
    LPCOLESTR,  # [in] lpszCaption
    ULONG,  # [in] cObjects
    POINTER(LPUNKNOWN),  # [in] ppUnk
    ULONG,  # [in] cPages
    LPCLSID,  # [in] pPageClsID
    LCID,  # [in] lcid
    DWORD,  # [in] dwReserved
    LPVOID,  # [in] pvReserved
)
